import java.util.*;
import javax.bluetooth.*;
import javax.bluetooth.UUID;

// Used to find nearby bluetooth devices.
public class BluetoothDeviceFinder implements DiscoveryListener, Runnable {
	private static final UUID uuid = new UUID("1101", true);	// Serial port uuid

	private Vector<RemoteDevice> 	devices;			// Devices found
	private Vector<ServiceRecord> 	records;			// Service records for found devices
	private Vector<String> 			urls;				// URLs of found devices
	private Vector<String> 			addresses;			// Bluetooth addresses of found devices
	private boolean					scanCompleted;		// False when a scan is being made

	// Scan for bluetooth devices and return a list of bluetooth URLs.
	public Vector<String> scanForBtDevice() {
		scanCompleted = false;

		devices = new Vector<RemoteDevice>();
		urls = new Vector<String>();
		addresses = new Vector<String>();

		// Allow the scan to be cancelled on a timeout
		new Thread(this).start();

		// Start scanning for devices
		synchronized(this) {
			try {
				LocalDevice.getLocalDevice().getDiscoveryAgent().startInquiry(DiscoveryAgent.GIAC, this);
				try {
					wait();
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			} catch (BluetoothStateException e) {
				e.printStackTrace();
			}
		}

		for (Enumeration<RemoteDevice> enum_d = devices.elements(); enum_d.hasMoreElements();) {
			RemoteDevice d = (RemoteDevice) enum_d.nextElement();
			System.out.println("[BLUETOOTH LOG] discovered address: " + d.getBluetoothAddress());

			synchronized (this) {
				records = new Vector<ServiceRecord>();
				try {
					LocalDevice.getLocalDevice().getDiscoveryAgent().searchServices(
							new int[] { 0x0100, 0x0101, 0x0A0, 0x0A1, 0x0A2, 0x0A3, 0x0A4 },
							new UUID[] { uuid }, d, this);
					try {
						wait();
					} catch (InterruptedException e) {
						e.printStackTrace();
					}

				} catch (BluetoothStateException e) {
					e.printStackTrace();
				}
			}

			for (Enumeration<ServiceRecord> enum_r = records.elements(); enum_r.hasMoreElements();) {
				ServiceRecord r = (ServiceRecord) enum_r.nextElement();

				String name = null;
				DataElement nameDataElement = r.getAttributeValue(0x0100);
				if (nameDataElement != null) {
					name = nameDataElement.getValue().toString();
				}
				System.out.println("[BLUETOOTH LOG] Name attribute: " + name);

				String url = r.getConnectionURL(ServiceRecord.NOAUTHENTICATE_NOENCRYPT, false);
				System.out.println("[BLUETOOTH LOG] url: " + url);
				urls.add(url);
				addresses.add(d.getBluetoothAddress() + " (" + name + ")");
			}
		}
		return addresses;
	}

	// Used to allow for a timeout after 30 seconds.
	public void run() {
		try {
			Thread.sleep(30000);	// 30 second timer
		} catch (InterruptedException e) {
			e.printStackTrace();
		}

		if (!scanCompleted) {
			System.out.println("[BLUETOOTH LOG] Scanning timed out");

			try {
				if (LocalDevice.getLocalDevice().getDiscoveryAgent().cancelInquiry(this)) {
					System.out.println("[BLUETOOTH LOG] Scan cancelled");
				}
			} catch (BluetoothStateException bse) {
				System.out.println("[BLUETOOTH LOG] Got BluetoothStateException: " + bse);
			}
		}
	}

	public void deviceDiscovered(RemoteDevice btDevice, DeviceClass cod) {
		devices.addElement(btDevice);
	}

	public synchronized void inquiryCompleted(int discType) {
		scanCompleted = true;
		notifyAll();
	}

	public void servicesDiscovered(int transID, ServiceRecord[] servRecord) {
		for (int i=0; i<servRecord.length; i++)
			records.addElement(servRecord[i]);
	}

	public synchronized void serviceSearchCompleted(int transID, int respCode) {
		notifyAll();
	}
}
